from pydantic_settings import BaseSettings
from pydantic import Field
from urllib.parse import quote


class Settings(BaseSettings):
    DATABASE_DRIVER: str = ''
    DATABASE_USER_: str = Field(alias='DATABASE_USER', default='')
    DATABASE_USER_DOMAIN: str = ''
    DATABASE_PASSWORD: str = ''
    DATABASE_HOST: str = ''
    DATABASE_PORT: str = ''
    DATABASE_NAME: str = ''
    VERSION_TABLE: str = ''

    SOURCE: str = ''
    BATCH_MODE: bool = False
    VERBOSITY: int = 0
    EDITOR: str = ''
    POST_CREATE_COMMAND: str = ''
    PREFIX: str = ''

    @property
    def DATABASE_USER(self) -> str:
        if self.DATABASE_USER_DOMAIN:
            return f'{self.DATABASE_USER_DOMAIN}\\{self.DATABASE_USER_}'
        return self.DATABASE_USER_

    @property
    def DATABASE(self) -> str:
        return (
            f'{self.DATABASE_DRIVER}://{self.DATABASE_USER}'
            f"{':' if self.DATABASE_PASSWORD else ''}"
            f'{quote(self.DATABASE_PASSWORD)}' 
            f'@{self.DATABASE_HOST}'
            f"{':' if self.DATABASE_PORT else ''}{self.DATABASE_PORT}"
            f'/{self.DATABASE_NAME}'
        )

    @property
    def sources_list(self) -> list[str]:
        return self.SOURCE.split()

    class Config:
        env_file = '.env'
        env_file_encoding = 'utf-8'
        extra = 'allow'



# Copyright 2022 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import logging
import sys

logger = logging.getLogger("classic.migrations")


def install_argparsers(global_parser, subparsers):
    parser = subparsers.add_parser(
        "init", parents=[global_parser], help="Initialize a new project"
    )
    parser.set_defaults(func=init)
    parser.add_argument("sources", nargs=1, help="Directory for migration scripts")
    parser.add_argument(
        "-d",
        "--database",
        help="Database, eg 'sqlite:///path/to/sqlite.db' "
        "or 'postgresql://user@host/db'",
    )


def init(args, config) -> int:
    migrations_path = pathlib.Path(args.sources[0]).resolve()
    if not migrations_path.exists():
        migrations_path.mkdir(exist_ok=True, parents=True)
    print(f"Created migrations directory {migrations_path}")
    return 0



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Mapping
from datetime import datetime
from contextlib import contextmanager
from importlib import import_module
from itertools import count
from logging import getLogger
from typing import Dict
from importlib_metadata import entry_points

import getpass
import os
import pickle
import socket
import time
import uuid

from classic.migrations import exceptions
from classic.migrations import utils
from classic.migrations.migrations import topological_sort, get_migration_hash

logger = getLogger("classic.migrations")


class TransactionManager:
    """
    Returned by the :meth:`~classic.migrations.backends.DatabaseBackend.transaction`
    context manager.

    If rollback is called, the transaction is flagged to be rolled back
    when the context manager block closes
    """

    def __init__(self, backend, rollback_on_exit=False):
        self.backend = backend
        self.rollback_on_exit = rollback_on_exit

    def __enter__(self):
        self._do_begin()
        return self

    def __exit__(self, exc_type, value, traceback):
        if exc_type:
            self._do_rollback()
            return None

        if self.rollback_on_exit:
            self._do_rollback()
        else:
            self._do_commit()

    def _do_begin(self):
        """
        Instruct the backend to begin a transaction
        """
        self.backend.begin()

    def _do_commit(self):
        """
        Instruct the backend to commit the transaction
        """
        self.backend.commit()

    def _do_rollback(self):
        """
        Instruct the backend to roll back the transaction
        """
        self.backend.rollback()


class SavepointTransactionManager(TransactionManager):

    id = None
    id_generator = count(1)

    def _do_begin(self):
        assert self.id is None
        self.id = "sp_{}".format(next(self.id_generator))
        self.backend.savepoint(self.id)

    def _do_commit(self):
        """
        This does nothing.

        Trying to the release savepoint here could cause an database error in
        databases where DDL queries cause the transaction to be committed
        and all savepoints released.
        """

    def _do_rollback(self):
        self.backend.savepoint_rollback(self.id)


class DatabaseBackend:

    driver_module = ""

    log_table = "history"
    lock_table = "lock"
    list_tables_sql = "SELECT table_name FROM information_schema.tables"
    migration_table = "versions"
    is_applied_sql = (
        "SELECT COUNT(1) FROM {0.migrations_schema_name_quoted}.{0.migration_table_quoted}"
        "WHERE id=:id"
    )
    mark_migration_sql = (
        "INSERT INTO {0.migrations_schema_name_quoted}.{0.migration_table_quoted} "
        "(migration_hash, migration_id, applied_at_utc) "
        "VALUES (:migration_hash, :migration_id, :when)"
    )
    unmark_migration_sql = (
        "DELETE FROM {0.migrations_schema_name_quoted}.{0.migration_table_quoted} WHERE "
        "migration_hash = :migration_hash"
    )
    applied_migrations_sql = (
        "SELECT migration_hash FROM "
        "{0.migrations_schema_name_quoted}.{0.migration_table_quoted} "
        "ORDER by applied_at_utc"
    )
    create_test_table_sql = "CREATE TABLE {table_name_quoted} " "(id INT PRIMARY KEY)"
    log_migration_sql = (
        "INSERT INTO {0.migrations_schema_name_quoted}.{0.log_table_quoted} "
        "(id, migration_hash, migration_id, operation, "
        "username, hostname, created_at_utc, comment) "
        "VALUES (:id, :migration_hash, :migration_id, "
        ":operation, :username, :hostname, :created_at_utc, :comment)"
    )
    create_lock_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.lock_table_quoted} ("
        "locked INT DEFAULT 1, "
        "ctime TIMESTAMP,"
        "pid INT NOT NULL,"
        "PRIMARY KEY (locked))"
    )
    insert_log_table_sql = (
        "INSERT INTO {0.migrations_schema_name_quoted}.{0.log_table_quoted} "
        "(id, migration_hash, migration_id, operation, created_at_utc, "
        "username, hostname, comment) "
        "VALUES "
        "(:id, :migration_hash, :migration_id, 'apply', :created_at_utc, "
        ":username, :hostname, :comment)"
    )
    insert_migration_table_from_log_table_sql = (
        "INSERT INTO {0.migrations_schema_name_quoted}.{0.migration_table_quoted} "
        "SELECT migration_hash, migration_id, created_at_utc "
        "FROM {0.migrations_schema_name_quoted}.{0.log_table_quoted}"
    )
    create_migration_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.migration_table_quoted} ( "
        # sha256 hash of the migration id
        "migration_hash VARCHAR(64), "
        # The migration id (ie path basename without extension)
        "migration_id VARCHAR(255), "
        # When this id was applied
        "applied_at_utc TIMESTAMP, "
        "PRIMARY KEY (migration_hash))"
    )
    create_log_table_sql = (
        "CREATE TABLE {0.migrations_schema_name_quoted}.{0.log_table_quoted} ( "
        "id VARCHAR(36), "
        "migration_hash VARCHAR(64), "
        "migration_id VARCHAR(255), "
        "operation VARCHAR(10), "
        "username VARCHAR(255), "
        "hostname VARCHAR(255), "
        "comment VARCHAR(255), "
        "created_at_utc TIMESTAMP, "
        "PRIMARY KEY (id))"
    )
    migrations_schema_exists_sql = ""
    create_migrations_schema_sql = "CREATE SCHEMA {0.migrations_schema_name_quoted};"
    _driver = None
    _is_locked = False
    _in_transaction = False
    _internal_schema_updated = False
    _transactional_ddl_cache: Dict[bytes, bool] = {}
    migrations_schema_name = "migrations"

    def __init__(self, dburi, migration_table):
        self.uri = dburi
        self.DatabaseError = self.driver.DatabaseError
        self._connection = self.connect(dburi)
        self.init_connection(self._connection)
        self.migration_table = migration_table
        self.has_transactional_ddl = self._transactional_ddl_cache.get(
            pickle.dumps(self.uri), True
        )

    def init_database(self):
        self._create_migrations_schema()
        self.create_lock_table()
        self.has_transactional_ddl = self._check_transactional_ddl()
        self._transactional_ddl_cache[
            pickle.dumps(self.uri)
        ] = self.has_transactional_ddl
        if self.needs_upgrading():
            self._upgrade()

    def needs_upgrading(self):
        list_tables = self.list_tables()
        return (self.log_table not in list_tables
            or self.lock_table not in list_tables
            or self.migration_table not in list_tables
        )

    def _upgrade(self):
        self._create_log_table()
        self._create_migration_table()
        cursor = self.execute(
            "SELECT migration_id, applied_at_utc "
            f"FROM {self.migrations_schema_name_quoted}.{self.migration_table_quoted}"
        )
        migration_id = ""
        created_at = datetime(1970, 1, 1)
        for migration_id, created_at in iter(cursor.fetchone,
                                             None):  # type: ignore
            migration_hash = get_migration_hash(migration_id)
            log_data = dict(
                self.get_log_data(),
                operation="apply",
                comment=(
                    "this log entry created automatically by an internal schema upgrade"
                ),
                created_at_utc=created_at,
                migration_hash=migration_hash,
                migration_id=migration_id,
            )
            self.execute(
                self.insert_log_table_sql.format(self),
                log_data,
            )

        self.execute(f"DROP TABLE {self.migrations_schema_name_quoted}.{self.migration_table_quoted}")
        self._create_migration_table()
        self.execute(
            self.insert_migration_table_from_log_table_sql.format(self)
        )

    def _create_migration_table(self):
        if self.migration_table not in self.list_tables():
            self.execute(self.create_migration_table_sql.format(self))

    def _create_log_table(self):
        if self.log_table not in self.list_tables():
            self.execute(self.create_log_table_sql.format(self))

    def _create_migrations_schema(self):
        try:
            self.execute(self.create_migrations_schema_sql.format(self))
        except self.DatabaseError:
             pass

    def _load_driver_module(self):
        """
        Load the dbapi driver module and register the base exception class
        """
        driver = get_dbapi_module(self.driver_module)
        exceptions.register(driver.DatabaseError)
        return driver

    @property
    def driver(self):
        if self._driver:
            return self._driver
        self._driver = self._load_driver_module()
        return self._driver

    @property
    def connection(self):
        return self._connection

    def init_connection(self, connection):
        """
        Called when creating a connection or after a rollback. May do any
        db specific tasks required to make the connection ready for use.
        """

    def copy(self):
        """
        Return a copy of the backend with a independent db
        connection.
        """
        return self.__class__(self.uri, self.migration_table)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.connection.close()

    def __getattr__(self, attrname):
        if attrname.endswith("_quoted"):
            unquoted = getattr(self, attrname.rsplit("_quoted")[0])
            return self.quote_identifier(unquoted)
        raise AttributeError(attrname)

    def connect(self, dburi):
        raise NotImplementedError()

    def quote_identifier(self, s):
        assert "\x00" not in s
        quoted = s.replace('"', '""')
        return f'"{quoted}"'

    def _check_transactional_ddl(self):
        """
        Return True if the database supports committing/rolling back
        DDL statements within a transaction
        """
        table_name_quoted = (self.migrations_schema_name_quoted+"."+
                      self.quote_identifier("migration_tmp_{}".format(utils.get_random_string(10))))
        sql = self.create_test_table_sql.format(table_name_quoted=table_name_quoted)
        try:
            with self.transaction(rollback_on_exit=True):
                self.execute(sql)
        except self.DatabaseError:
            return False

        try:
            with self.transaction():
                self.execute(f"DROP TABLE {table_name_quoted}")
        except self.DatabaseError:
            return True
        return False

    def list_tables(self, **kwargs):
        """
        Return a list of tables present in the backend.
        This is used by the test suite to clean up tables
        generated during testing
        """
        cursor = self.execute(
            self.list_tables_sql.format(self),
            dict({"database": self.uri.database}, **kwargs),
        )
        return [row[0] for row in cursor.fetchall()]

    def transaction(self, rollback_on_exit=False):
        if not self._in_transaction:
            return TransactionManager(self, rollback_on_exit=rollback_on_exit)
        else:
            return SavepointTransactionManager(self, rollback_on_exit=rollback_on_exit)

    def cursor(self):
        return self.connection.cursor()

    def commit(self):
        try:
            self.connection.commit()
        except Exception as e:
            print(str(e))
        self._in_transaction = False

    def rollback(self):
        self.connection.rollback()
        self.init_connection(self.connection)
        self._in_transaction = False

    def begin(self):
        """
        Begin a new transaction
        """
        assert not self._in_transaction
        self._in_transaction = True
        self.execute("BEGIN")

    def savepoint(self, id):
        """
        Create a new savepoint with the given id
        """
        self.execute("SAVEPOINT {}".format(id))

    def savepoint_release(self, id):
        """
        Release (commit) the savepoint with the given id
        """
        self.execute("RELEASE SAVEPOINT {}".format(id))

    def savepoint_rollback(self, id):
        """
        Rollback the savepoint with the given id
        """
        self.execute("ROLLBACK TO SAVEPOINT {}".format(id))

    @contextmanager
    def disable_transactions(self):
        """
        Disable the connection's transaction support, for example by
        setting the isolation mode to 'autocommit'
        """
        self.rollback()
        yield

    @contextmanager
    def lock(self, timeout=10):
        """
        Create a lock to prevent concurrent migrations.

        :param timeout: duration in seconds before raising a LockTimeout error.
        """
        if self._is_locked:
            yield
            return

        pid = os.getpid()
        self._insert_lock_row(pid, timeout)
        try:
            self._is_locked = True
            yield
            self._is_locked = False
        finally:
            self._delete_lock_row(pid)

    def _insert_lock_row(self, pid, timeout, poll_interval=0.5):
        poll_interval = min(poll_interval, timeout)
        started = time.time()
        while True:
            try:
                with self.transaction():
                    self.execute(
                        "INSERT INTO " +
                        self.migrations_schema_name_quoted.format(self) + "." +
                        self.lock_table_quoted+
                        " (locked, ctime, pid) " +
                        "VALUES (1, :when, :pid)",
                        {"when": datetime.utcnow(), "pid": pid},
                    )
            except self.DatabaseError:
                if timeout and time.time() > started + timeout:
                    cursor = self.execute(
                        "SELECT pid FROM "+
                        self.migrations_schema_name_quoted.format(self)+
                        ".{}".format(self.lock_table_quoted)
                    )
                    row = cursor.fetchone()
                    if row:
                        raise exceptions.LockTimeout(
                            "Process {} has locked this database "
                            "(run yoyo break-lock to remove this lock)".format(row[0])
                        )
                    else:
                        raise exceptions.LockTimeout(
                            "Database locked "
                            "(run yoyo break-lock to remove this lock)"
                        )
                time.sleep(poll_interval)
            else:
                return

    def _delete_lock_row(self, pid):
        with self.transaction():
            self.execute(
                f"DELETE FROM {self.migrations_schema_name_quoted.format(self)}.{self.lock_table_quoted} WHERE pid=:pid;",
                {"pid": pid},
            )

    def break_lock(self):
        with self.transaction():
            self.execute("DELETE FROM "+self.migrations_schema_name_quoted.format(self)+".{}".format(self.lock_table_quoted))

    def execute(self, sql, params=None):
        """
        Create a new cursor, execute a single statement and return the cursor
        object.

        :param sql: A single SQL statement, optionally with named parameters
                    (eg 'SELECT * FROM foo WHERE :bar IS NULL')
        :param params: A dictionary of parameters
        """
        if params and not isinstance(params, Mapping):
            raise TypeError("Expected dict or other mapping object")

        cursor = self.cursor()
        sql, params = utils.change_param_style(self.driver.paramstyle, sql, params)
        cursor.execute(sql, params)
        return cursor

    def create_lock_table(self):
        """
        Create the lock table if it does not already exist.
        """
        if self.lock_table not in self.list_tables():
                self.execute(self.create_lock_table_sql.format(self))
                self.commit()

    def ensure_internal_schema_updated(self):
        """
        Check and upgrade yoyo's internal schema.
        """
        if self._internal_schema_updated:
            return
        if self.needs_upgrading():
            assert not self._in_transaction
            with self.lock():
                self._upgrade()
                self.connection.commit()
                self._internal_schema_updated = True

    def is_applied(self, migration):
        return migration.hash in self.get_applied_migration_hashes()

    def get_applied_migration_hashes(self):
        """
        Return the list of migration hashes in the order in which they
        were applied
        """
        self.ensure_internal_schema_updated()
        sql = self.applied_migrations_sql.format(self)
        return [row[0] for row in self.execute(sql).fetchall()]

    def to_apply(self, migrations):
        """
        Return the subset of migrations not already applied.
        """
        applied = self.get_applied_migration_hashes()
        ms = (m for m in migrations if m.hash not in applied)
        return migrations.__class__(topological_sort(ms), migrations.post_apply)

    def to_rollback(self, migrations):
        """
        Return the subset of migrations already applied and which may be
        rolled back.

        The order of migrations will be reversed.
        """
        applied = self.get_applied_migration_hashes()
        ms = (m for m in migrations if m.hash in applied)
        return migrations.__class__(
            reversed(list(topological_sort(ms))), migrations.post_apply
        )

    def apply_migrations(self, migrations, force=False):
        if migrations:
            self.apply_migrations_only(migrations, force=force)
            self.run_post_apply(migrations, force=force)

    def apply_migrations_only(self, migrations, force=False):
        """
        Apply the list of migrations, but do not run any post-apply hooks
        present.
        """
        if not migrations:
            return
        for m in migrations:
            try:
                self.apply_one(m, force=force)
            except exceptions.BadMigration:
                continue

    def run_post_apply(self, migrations, force=False):
        """
        Run any post-apply migrations present in ``migrations``
        """
        for m in migrations.post_apply:
            self.apply_one(m, mark=False, force=force)

    def rollback_migrations(self, migrations, force=False):
        self.ensure_internal_schema_updated()
        if not migrations:
            return
        for m in migrations:
            try:
                self.rollback_one(m, force)
            except exceptions.BadMigration:
                continue

    def mark_migrations(self, migrations):
        self.ensure_internal_schema_updated()
        with self.transaction():
            for m in migrations:
                try:
                    self.mark_one(m)
                except exceptions.BadMigration:
                    continue

    def unmark_migrations(self, migrations):
        self.ensure_internal_schema_updated()
        with self.transaction():
            for m in migrations:
                try:
                    self.unmark_one(m)
                except exceptions.BadMigration:
                    continue

    def apply_one(self, migration, force=False, mark=True):
        """
        Apply a single migration
        """
        logger.info("Applying %s", migration.id)
        self.ensure_internal_schema_updated()
        with self.copy() as migration_backend:
            migration.process_steps(migration_backend, "apply", force=force)
        self.log_migration(migration, "apply")
        if mark:
            with self.transaction():
                self.mark_one(migration, log=False)

    def rollback_one(self, migration, force=False):
        """
        Rollback a single migration
        """
        logger.info("Rolling back %s", migration.id)
        self.ensure_internal_schema_updated()
        with self.copy() as migration_backend:
            migration.process_steps(migration_backend, "rollback", force=force)
        self.log_migration(migration, "rollback")
        with self.transaction():
            self.unmark_one(migration, log=False)

    def unmark_one(self, migration, log=True):
        self.ensure_internal_schema_updated()
        sql = self.unmark_migration_sql.format(self)
        self.execute(sql, {"migration_hash": migration.hash})
        if log:
            self.log_migration(migration, "unmark")

    def mark_one(self, migration, log=True):
        self.ensure_internal_schema_updated()
        logger.info("Marking %s applied", migration.id)
        sql = self.mark_migration_sql.format(self)
        self.execute(
            sql,
            {
                "migration_hash": migration.hash,
                "migration_id": migration.id,
                "when": datetime.utcnow(),
            },
        )
        if log:
            self.log_migration(migration, "mark")

    def log_migration(self, migration, operation, comment=None):
        sql = self.log_migration_sql.format(self)
        self.execute(sql, self.get_log_data(migration, operation, comment))

    def get_log_data(self, migration=None, operation="apply", comment=None):
        """
        Return a dict of data for insertion into the ``_yoyo_log`` table
        """
        assert operation in {"apply", "rollback", "mark", "unmark"}
        return {
            "id": str(uuid.uuid1()),
            "migration_id": migration.id if migration else None,
            "migration_hash": migration.hash if migration else None,
            "username": getpass.getuser(),
            "hostname": socket.getfqdn(),
            "created_at_utc": datetime.utcnow(),
            "operation": operation,
            "comment": (comment if comment else
                            (migration.module.inside_sql_comment
                             if migration.module.inside_sql_comment else None
                            )
                        ),
        }


def get_backend_class(name):
    backend_eps = entry_points(group="migrations.backends")
    return backend_eps[name].load()


def get_dbapi_module(name):
    """
    Import and return the named DB-API driver module
    """
    return import_module(name)



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from classic.migrations.backends.base import DatabaseBackend


class MySQLBackend(DatabaseBackend):

    driver_module = "pymysql"
    list_tables_sql = (
        "SELECT table_name FROM information_schema.tables "
        "WHERE table_schema = :database"
    )

    def connect(self, dburi):
        kwargs = {"db": dburi.database}
        kwargs.update(dburi.args)
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["passwd"] = dburi.password
        if dburi.hostname is not None:
            kwargs["host"] = dburi.hostname
        if dburi.port is not None:
            kwargs["port"] = dburi.port
        if "unix_socket" in dburi.args:
            kwargs["unix_socket"] = dburi.args["unix_socket"]
        if "ssl" in dburi.args:
            kwargs["ssl"] = {}

            if "sslca" in dburi.args:
                kwargs["ssl"]["ca"] = dburi.args["sslca"]

            if "sslcapath" in dburi.args:
                kwargs["ssl"]["capath"] = dburi.args["sslcapath"]

            if "sslcert" in dburi.args:
                kwargs["ssl"]["cert"] = dburi.args["sslcert"]

            if "sslkey" in dburi.args:
                kwargs["ssl"]["key"] = dburi.args["sslkey"]

            if "sslcipher" in dburi.args:
                kwargs["ssl"]["cipher"] = dburi.args["sslcipher"]

        kwargs["db"] = dburi.database
        return self.driver.connect(**kwargs)

    def quote_identifier(self, identifier):
        sql_mode = self.execute("SHOW VARIABLES LIKE 'sql_mode'").fetchone()[1]
        if "ansi_quotes" in sql_mode.lower():
            return super(MySQLBackend, self).quote_identifier(identifier)
        return "`{}`".format(identifier)


class MySQLdbBackend(MySQLBackend):
    driver_module = "MySQLdb"



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from classic.migrations.backends.base import DatabaseBackend


class SnowflakeBackend(DatabaseBackend):

    driver_module = "snowflake.connector"

    def connect(self, dburi):
        database, schema = dburi.database.split("/")
        return self.driver.connect(
            user=dburi.username,
            password=dburi.password,
            account=dburi.hostname,
            database=database,
            schema=schema,
            **dburi.args,
        )

    def savepoint(self, id):
        pass

    def savepoint_release(self, id):
        pass

    def savepoint_rollback(self, id):
        pass



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple
from urllib.parse import parse_qsl
from urllib.parse import quote
from urllib.parse import unquote
from urllib.parse import urlencode
from urllib.parse import urlsplit
from urllib.parse import urlunsplit

from classic.migrations.backends import get_backend_class
from classic.migrations.migrations import default_migration_table


_DatabaseURI = namedtuple(
    "_DatabaseURI", "scheme username password hostname port database args"
)


class DatabaseURI(_DatabaseURI):
    @property
    def netloc(self):
        hostname = self.hostname or ""
        if self.port:
            hostpart = "{}:{}".format(hostname, self.port)
        else:
            hostpart = hostname

        if self.username:
            return "{}:{}@{}".format(
                quote(self.username, safe=""),
                quote(self.password or "", safe=""),
                hostpart,
            )
        else:
            return hostpart

    def __str__(self):
        return urlunsplit(
            (self.scheme, self.netloc, self.database, urlencode(self.args), "")
        )

    @property
    def uri(self):
        return str(self)


class BadConnectionURI(Exception):
    """
    An invalid connection URI
    """


def get_backend(uri, migration_table=default_migration_table):
    """
    Connect to the given DB uri in the format
    ``driver://user:pass@host:port/database_name?param=value``,
    returning a :class:`DatabaseBackend` object
    """
    parsed = parse_uri(uri)
    try:
        backend_class = get_backend_class(parsed.scheme.lower())
    except KeyError:
        raise BadConnectionURI(
            "Unrecognised database connection scheme %r" % parsed.scheme
        )
    backend = backend_class(parsed, migration_table)
    backend.init_database()
    return backend


def parse_uri(s):
    """
    Examples::

        >>> parse_uri('postgres://fred:bassett@server:5432/fredsdatabase')
        ('postgres', 'fred', 'bassett', 'server', 5432, 'fredsdatabase', None)
        >>> parse_uri('mysql:///jimsdatabase')
        ('mysql', None, None, None, None, 'jimsdatabase', None, None)
        >>> parse_uri('odbc://user:password@server/database?DSN=dsn')
        ('odbc', 'user', 'password', 'server', None, 'database', {'DSN':'dsn'})
    """
    result = urlsplit(s)

    if not result.scheme:
        raise BadConnectionURI("No scheme specified in connection URI %r" % s)

    return DatabaseURI(
        scheme=result.scheme,
        username=(unquote(result.username) if result.username is not None else None),
        password=(unquote(result.password) if result.password is not None else None),
        hostname=result.hostname,
        port=result.port,
        database=result.path[1:] if result.path else None,
        args=dict(parse_qsl(result.query)),
    )



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from getpass import getpass
import argparse
import configparser
import logging
import os
import sys
import typing as t

from classic.migrations import connections
from classic.migrations import default_migration_table
from classic.migrations import logger
from classic.migrations import utils
from classic.migrations.settings import Settings

settings = Settings()

verbosity_levels = {
    0: logging.ERROR,
    1: logging.WARN,
    2: logging.INFO,
    3: logging.DEBUG,
}

min_verbosity = min(verbosity_levels)
max_verbosity = max(verbosity_levels)

LEGACY_CONFIG_FILENAME = ".yoyo-migrate"


class InvalidArgument(Exception):
    pass


def parse_args(
    argv=None,
) -> t.Tuple[argparse.ArgumentParser, argparse.Namespace]:
    """
    Parse the config file and command line args.

    :return: tuple of ``(parsed config file, argument parser, parsed arguments)``
    """
    #: List of arguments whose defaults should be read from the config file
    config_args = {
        "batch_mode": "getboolean",
        "sources": "get",
        "database": "get",
        "verbosity": "getint",
        "migration_table": "get",
    }

    globalparser, argparser, subparsers = make_argparser()

    # Initial parse to extract --config and any global arguments
    global_args, _ = globalparser.parse_known_args(argv)

    # Now parse for real, starting from the top
    args = argparser.parse_args(argv)
    # Update the args namespace with the global args.
    # This ensures that global args (eg '-v) are recognized regardless
    # of whether they were placed before or after the subparser command.
    # If we do not do this then the sub_parser copy of the argument takes
    # precedence, and overwrites any global args set before the command name.
    args.__dict__.update(globalparser.parse_known_args(argv)[0].__dict__)

    return argparser, args


def make_argparser():
    """
    Return a top-level ArgumentParser parser object,
    plus a list of sub_parsers
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "--config", "-c", default=None, help="Path to config file"
    )
    global_parser.add_argument(
        "-v",
        dest="verbosity",
        action="count",
        default=min_verbosity,
        help="Verbose output. Use multiple times to increase level of verbosity",
    )
    global_parser.add_argument(
        "-b",
        "--batch",
        dest="batch_mode",
        action="store_true",
        default=(not sys.stdout.isatty()),
        help="Run in batch mode" ". Turns off all user prompts",
    )

    # global_parser.add_argument(
    #     "-s",
    #     "--source",
    #     dest="sources",
    #     help="Source directory of migration scripts",
    # )

    global_parser.add_argument(
        "--no-config-file",
        "--no-cache",
        dest="use_config_file",
        action="store_false",
        default=True,
        help="Don't look for a migrations.ini config file",
    )
    argparser = argparse.ArgumentParser(prog="migrations", parents=[global_parser])

    subparsers = argparser.add_subparsers(help="Commands help")

    from . import migrate
    from . import newmigration
    from . import init

    init.install_argparsers(global_parser, subparsers)
    migrate.install_argparsers(global_parser, subparsers)
    newmigration.install_argparsers(global_parser, subparsers)

    return global_parser, argparser, subparsers


def configure_logging(level):
    """
    Configure the python logging module with the requested loglevel
    """
    logging.basicConfig(level=verbosity_levels[level])


def prompt_save_config(config, path):
    # Offer to save the current configuration for future runs
    # Don't cache anything in batch mode (because we can't prompt to find the
    # user's preference).

    # if utils.confirm(
    #     "Save migration configuration to {}?\n"
    #     "This is saved in plain text and "
    #     "contains your database password.\n\n"
    #     "Answering 'y' means you do not have to specify "
    #     "the migration source or database connection "
    #     "for future runs".format(path)
    # ):
    #     save_config(config, path)
    ...

def upgrade_legacy_config(args, config, sources):

    for dir in reversed(sources):
        path = os.path.join(dir, LEGACY_CONFIG_FILENAME)
        if not os.path.isfile(path):
            continue

        legacy_config = read_config(path)

        def transfer_setting(oldname, newname, transform=None, section="DEFAULT"):
            try:
                config.get(section, newname)
            except configparser.NoOptionError:
                try:
                    value = legacy_config.get(section, oldname)
                except configparser.NoOptionError:
                    pass
                else:
                    if transform:
                        value = transform(value)
                    config.set(section, newname, value)

        transfer_setting("dburi", "database")
        transfer_setting(
            "migration_table",
            "migration_table",
            lambda v: (default_migration_table if v == "None" else v),
        )

        config_path = args.config or CONFIG_FILENAME
        if not args.batch_mode:
            if utils.confirm(
                "Move legacy configuration in {!r} to {!r}?".format(path, config_path)
            ):
                save_config(config, config_path)
                try:
                    if utils.confirm(
                        "Delete legacy configuration file {!r}".format(path)
                    ):
                        os.unlink(path)
                except OSError:
                    logger.warn(
                        "Could not remove %r. Manually remove this file "
                        "to avoid future warnings",
                        path,
                    )
                return True
        else:
            logger.warn(
                "Found legacy configuration in %r. Run "
                "yoyo in interactive mode to update your "
                "configuration files",
                path,
            )

            try:
                args.database = args.database or legacy_config.get("DEFAULT", "dburi")
            except configparser.NoOptionError:
                pass
            try:
                if not vars(args).get("migration_table"):
                    args.migration_table = legacy_config.get(
                        "DEFAULT", "migration_table"
                    )
            except configparser.NoOptionError:
                pass

        return False



def get_backend(args, settings:Settings):


    dburi = getattr(args, 'database', settings.DATABASE)
    dburi = dburi or settings.DATABASE
    if dburi is None:
        raise InvalidArgument("Please specify a database uri")
    migration_table = default_migration_table
    try:
        migration_table = args.migration_table
    except AttributeError:
        pass

    if len(migration_table)==0:
        migration_table = (
            settings.VERSION_TABLE
            if len(settings.VERSION_TABLE) == 0
            else default_migration_table
        )

    try:
        if args.prompt_password:
            password = getpass("Password for %s: " % dburi)
            parsed = connections.parse_uri(dburi)
            dburi = parsed._replace(password=password).uri
    except AttributeError:
        pass

    return connections.get_backend(dburi, migration_table)


def main(argv=None):

    argparser, args = parse_args(argv)
    if getattr(args, "func", None) is None:
        argparser.print_usage(sys.stderr)
        argparser.exit(1)

    args.batch_mode = settings.BATCH_MODE
    verbosity = args.verbosity
    verbosity = min(max_verbosity, max(min_verbosity, verbosity))
    configure_logging(verbosity)

    try:
        if vars(args).get("func"):
            exitcode = args.func(args, settings) #config
    except InvalidArgument as e:
        argparser.error(e.args[0])

    return exitcode


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright 2015 Oliver Cope
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from contextlib import contextmanager

from classic.migrations.backends.base import DatabaseBackend


class PostgresqlBackend(DatabaseBackend):
    """
    Backend for PostgreSQL and PostgreSQL compatible databases.

    This backend uses psycopg2. See
    :class:`classic.migrations.backends.core.postgresql.PostgresqlPsycopgBackend`
    if you need psycopg3.
    """

    driver_module = "psycopg2"
    schema = None
    migrations_schema_exists_sql = "SELECT 1 FROM information_schema.schemata WHERE schema_name =  '{0.migrations_schema_name}';"
    list_tables_sql = "SELECT table_name FROM information_schema.tables where table_schema = '{0.migrations_schema_name}';"

    # list_tables_sql = (
    #     "SELECT table_name FROM information_schema.tables "
    #     "WHERE table_schema = :schema"
    # )

    @property
    def TRANSACTION_STATUS_IDLE(self):
        from psycopg2.extensions import TRANSACTION_STATUS_IDLE

        return TRANSACTION_STATUS_IDLE

    def connect(self, dburi):
        kwargs = {"dbname": dburi.database, "autocommit": True}

        # Default to autocommit mode: without this psycopg sends a BEGIN before
        # every query, causing a warning when we then explicitly start a
        # transaction. This warning becomes an error in CockroachDB. See
        # https://todo.sr.ht/~olly/yoyo/71
        kwargs["autocommit"] = True

        kwargs.update(dburi.args)
        if dburi.username is not None:
            kwargs["user"] = dburi.username
        if dburi.password is not None:
            kwargs["password"] = dburi.password
        if dburi.port is not None:
            kwargs["port"] = dburi.port
        if dburi.hostname is not None:
            kwargs["host"] = dburi.hostname
        self.schema = kwargs.pop("schema", None)
        autocommit = bool(kwargs.pop("autocommit"))
        connection = self.driver.connect(**kwargs)
        connection.autocommit = autocommit
        return connection

    @contextmanager
    def disable_transactions(self):
        with super(PostgresqlBackend, self).disable_transactions():
            saved = self.connection.autocommit
            self.connection.autocommit = True
            yield
            self.connection.autocommit = saved

    def init_connection(self, connection):
        if self.schema:
            cursor = connection.cursor()
            cursor.execute("SET search_path TO {}".format(self.schema))

    def list_tables(self):
        current_schema = self.execute("SELECT current_schema").fetchone()[0]
        return super(PostgresqlBackend, self).list_tables(schema=current_schema)

    def commit(self):
        # The connection is in autocommit mode and ignores calls to
        # ``commit()`` and ``rollback()``, so we have to issue the SQL directly
        self.execute("COMMIT")
        super().commit()

    def rollback(self):
        self.execute("ROLLBACK")
        super().rollback()

    def begin(self):
        if self.connection.info.transaction_status != self.TRANSACTION_STATUS_IDLE:
            warnings.warn(
                "Nested transaction requested; "
                "this will raise an exception in some "
                "PostgreSQL-compatible databases"
            )
        return super().begin()


class PostgresqlPsycopgBackend(PostgresqlBackend):
    """
    Like PostgresqlBackend, but using the newer Psycopg 3.
    """

    driver_module = "psycopg"

    @property
    def TRANSACTION_STATUS_IDLE(self):
        from psycopg.pq import TransactionStatus

        return TransactionStatus.IDLE



from classic.migrations.backends.core.mysql import MySQLBackend
from classic.migrations.backends.core.sqlite3 import SQLiteBackend
from classic.migrations.backends.core.postgresql import PostgresqlBackend
from classic.migrations.backends.core.postgresql import PostgresqlPsycopgBackend

__all__ = [
    "MySQLBackend",
    "SQLiteBackend",
    "PostgresqlBackend",
    "PostgresqlPsycopgBackend",
]



from typing import Set
from typing import Dict
from typing import Mapping
from typing import TypeVar
from typing import Iterable
from typing import Collection
from collections import defaultdict
from heapq import heappop
from heapq import heappush


class CycleError(ValueError):
    """
    Raised when cycles exist in the input graph.

    The second element in the args attribute of instances will contain the
    sequence of nodes in which the cycle lies.
    """


T = TypeVar("T")


def topological_sort(
    items: Iterable[T], dependency_graph: Mapping[T, Collection[T]]
) -> Iterable[T]:

    # Tag each item with its input order
    pqueue = list(enumerate(items))
    ordering = {item: ix for ix, item in pqueue}
    seen_since_last_change = 0
    output: Set[T] = set()

    # Map blockers to the list of items they block
    blocked_on: Dict[T, Set[T]] = defaultdict(set)
    blocked: Set[T] = set()

    while pqueue:
        if seen_since_last_change == len(pqueue) + len(blocked):
            raise_cycle_error(ordering, pqueue, blocked_on)

        _, n = heappop(pqueue)

        blockers = {
            d for d in dependency_graph.get(n, []) if d not in output and d in ordering
        }
        if not blockers:
            seen_since_last_change = 0
            output.add(n)
            if n in blocked:
                blocked.remove(n)
            yield n
            for b in blocked_on.pop(n, []):
                if not any(b in other for other in blocked_on.values()):
                    heappush(pqueue, (ordering[b], b))
        else:
            if n in blocked:
                seen_since_last_change += 1
            else:
                seen_since_last_change = 0
                blocked.add(n)
                for b in blockers:
                    blocked_on[b].add(n)
    if blocked_on:
        raise_cycle_error(ordering, pqueue, blocked_on)


def raise_cycle_error(ordering, pqueue, blocked_on):
    bad = next((item for item in blocked_on if item not in ordering), None)
    if bad:
        raise ValueError(f"Dependency graph contains a non-existent node {bad!r}")
    unresolved = {n for _, n in pqueue}
    unresolved.update(*blocked_on.values())
    if unresolved:
        raise CycleError(
            f"Dependency graph loop detected among {unresolved!r}",
            list(sorted(unresolved, key=ordering.get)),
        )
    raise AssertionError("raise_cycle_error called but no unresovled nodes exist")
